import os
from collections import OrderedDict

import numpy as np

from scipy.ndimage import rotate
from scipy.special import gammainc, gamma, gammaincinv

from astropy.convolution import convolve
from astropy.nddata import block_reduce
from astropy.modeling import FittableModel, Parameter, custom_model, models

__all__ = [
    'make_grid', 'PSFModel',
    'Nuker2D', 'Moffat2D', 'EllipMoffat2D',
    'sersic_enclosed', 'sersic_enclosed_inv', 'sersic_enclosed_model',
    'petrosian_profile', 'petrosian_model', 
]


def make_grid(size, factor=1):
    assert isinstance(factor, int)
    x_arange = y_arange = (np.arange(0.5, size * factor, 1) / factor) - 0.5
    return np.meshgrid(x_arange, y_arange)


class PSFModel(FittableModel):
    oversample = None

    _cache_grid = True
    _cached_grid_size = 0
    _cached_grid_factor = 0
    _cached_grid = None

    def clear_cached_grid(self):
        self._cached_grid_size = 0
        self._cached_grid_factor = 0
        if self._cached_grid is not None:
            del self._cached_grid
        self._cached_grid = None

    @property
    def cache_grid(self):
        return self._cache_grid

    @cache_grid.setter
    def cache_grid(self, value):
        if value == False:
            self._cache_grid = False
            self.clear_cached_grid()
        elif value == True:
            self._cache_grid = True
        else:
            raise ValueError("{} is not a bool, use True or False".format(value))

    def evaluate(self, *args, **kwargs):
        psf_p = args[-1]
        args = args[:-1]

        x = args[0]
        y = args[1]

        assert not np.any(x < 0), 'negative pixel values not supported at this time'
        assert not np.any(y < 0), 'negative pixel values not supported at this time'

        grid_size = max([i.max() + 1 for i in [x, y]])
        grid_factor = self.oversample if isinstance(self.oversample, int) else 1

        if grid_size == self._cached_grid_size and self._cached_grid_factor == grid_factor:
            main_grid = self._cached_grid
        else:
            main_grid = make_grid(grid_size, factor=grid_factor)

            if self.cache_grid:
                self._cached_grid = main_grid
                self._cached_grid_size = grid_size
                self._cached_grid_factor = grid_factor

        x_grid, y_grid = main_grid

        model_image = self._model.evaluate(x_grid, y_grid, *args[self.n_inputs:])

        if isinstance(self.oversample, int):
            model_image = block_reduce(model_image, grid_factor) / grid_factor ** 2

        elif isinstance(self.oversample, tuple):
            sub_grid_x0, sub_grid_y0, sub_grid_size, sub_grid_factor = self.oversample

            if isinstance(sub_grid_x0, str):
                assert sub_grid_x0 in self._model.param_names, "oversample param '{}' is not in the wrapped model param list".format(
                    sub_grid_x0)

                idx = self._model.param_names.index(sub_grid_x0)
                sub_grid_x0 = args[self.n_inputs:][idx][0]

            if isinstance(sub_grid_y0, str):
                assert sub_grid_y0 in self._model.param_names, "oversample param '{}' is not in the wrapped model param list".format(
                    sub_grid_y0)

                idx = self._model.param_names.index(sub_grid_y0)
                sub_grid_y0 = args[self.n_inputs:][idx][0]

            x_sub_grid, y_sub_grid = make_grid(sub_grid_size, factor=sub_grid_factor)

            x_sub_grid += int(sub_grid_x0) - sub_grid_size // 2
            y_sub_grid += int(sub_grid_y0) - sub_grid_size // 2

            sub_model_oversampled_image = self._model.evaluate(x_sub_grid, y_sub_grid, *args[self.n_inputs:])

            # Experimental  
            # over_sampled_sub_model_x0 = np.argmin(
            #     np.abs(x_sub_grid[0, :] - 1 / (2 * sub_grid_factor) - (sub_grid_x0 * sub_grid_factor)))
            # over_sampled_sub_model_y0 = np.argmin(
            #     np.abs(y_sub_grid[:, 0] - 1 / (2 * sub_grid_factor) - (sub_grid_y0 * sub_grid_factor)))
            #
            # sub_model_oversampled_image[
            #     over_sampled_sub_model_y0,
            #     over_sampled_sub_model_x0
            # ] = self._model.evaluate(sub_grid_x0, sub_grid_y0, *args[self.n_inputs:])

            sub_model_image = block_reduce(sub_model_oversampled_image, sub_grid_factor) / sub_grid_factor ** 2

            x_sub_min = int(x_sub_grid[0][0] - 1 / (2 * sub_grid_factor)) + 1
            y_sub_min = int(y_sub_grid[0][0] - 1 / (2 * sub_grid_factor)) + 1

            model_image[
            y_sub_min: y_sub_min + sub_grid_size,
            x_sub_min: x_sub_min + sub_grid_size
            ] = sub_model_image

        if self.psf is None:
            return model_image[y.astype(int), x.astype(int)]

        else:
            psf = self.psf
            if psf_p[0] != 0:
                psf = rotate(psf, psf_p[0], reshape=False)
            return convolve(model_image, psf)[y.astype(int), x.astype(int)]

    @property
    def model(self):

        model = self._model.copy()
        for param in model.param_names:
            setattr(model, param, getattr(self, param).value)

        fixed = self.fixed
        del fixed['psf_pa']

        bounds = self.bounds
        del bounds['psf_pa']

        model.fixed.update(fixed)
        model.bounds.update(bounds)

        return model

    @staticmethod
    def wrap(model, psf=None, oversample=None):
        if isinstance(model, PSFModel):
            raise TypeError("Can not wrap a PSFModel, try: PSFModel.wrap(psf_model.model)")

        # Extract model params
        params = OrderedDict(
            (param_name, Parameter(param_name, default=param_val)) for param_name, param_val in
            zip(model.param_names, model.parameters)
        )

        # Prepare class attributes
        members = OrderedDict([
            ('__module__', '__main__'),
            ('__doc__', 'PSF Wrapped Model\n{}'.format(model.__doc__)),
            ('n_inputs', model.n_inputs),
            ('n_outputs', model.n_outputs),
            ('psf', psf),
            ('_model', model),
        ])

        # Add params to class attributes
        members.update(params)
        members.update({'psf_pa': Parameter('psf_pa', default=0)})

        # Construct new model class
        model_class = type('PSFModel', (PSFModel,), members)

        # Init new model from new model class
        new_model = model_class()

        # Sync model states
        new_model.fixed.update(model.fixed)
        new_model.bounds.update(model.bounds)

        # add oversample regions
        if oversample is not None:
            if isinstance(oversample, int) or isinstance(oversample, tuple):
                new_model.oversample = oversample
            else:
                raise ValueError("oversample should be a single int factor or a tuple (x, y, size, factor).")

        # Return new model
        return new_model


@custom_model
def Nuker2D(x, y, amplitude=1, r_eff=1, x_0=0, y_0=0, a=1, b=2, g=0, ellip=0, theta=0):
    """Two dimensional Nuker2D model"""
    A, B = 1 * r_eff, (1 - ellip) * r_eff
    cos_theta, sin_theta = np.cos(theta), np.sin(theta)
    x_maj = (x - x_0) * cos_theta + (y - y_0) * sin_theta
    x_min = -(x - x_0) * sin_theta + (y - y_0) * cos_theta
    r = np.sqrt((x_maj / A) ** 2 + (x_min / B) ** 2)

    return 2 ** ((b - g) / a) * amplitude * (r_eff / r) ** (g) * (1 + (r / r_eff) ** a) ** ((g - b) / a)


@custom_model
def Moffat2D(x, y, amplitude=1.0, x_0=0.0, y_0=0.0, gamma=1.0, alpha=1.0):
    """Two dimensional Moffat function."""
    rr_gg = ((x - x_0) ** 2 + (y - y_0) ** 2) / gamma ** 2
    return amplitude * (1 + rr_gg) ** (-alpha)


@custom_model
def EllipMoffat2D(x, y, amplitude=1.0, x_0=0.0, y_0=0.0, gamma=1.0, alpha=1.0, ellip=0, theta=0, r=1):
    """Two dimensional Moffat function."""

    a, b = 1 * r, (1 - ellip) * r
    cos_theta, sin_theta = np.cos(theta), np.sin(theta)
    x_maj = (x - x_0) * cos_theta + (y - y_0) * sin_theta
    x_min = -(x - x_0) * sin_theta + (y - y_0) * cos_theta
    z = np.sqrt((x_maj / a) ** 2 + (x_min / b) ** 2)

    rr_gg = (z) / gamma ** 2

    return amplitude * (1 + rr_gg) ** (-alpha)


def sersic_enclosed(
        r,
        amplitude,
        r_eff,
        n,
):
    bn = gammaincinv(2. * n, 0.5)
    x = bn * (r / r_eff) ** (1 / n)
    g = gamma(2. * n) * gammainc(2. * n, x)

    return amplitude * (r_eff) ** 2 * 2 * np.pi * n * ((np.exp(bn)) / (bn) ** (2 * n)) * g


def sersic_enclosed_inv(
        f,
        amplitude,
        r_eff,
        n,
):
    bn = gammaincinv(2. * n, 0.5)
    g = f / (amplitude * (r_eff) ** 2 * 2 * np.pi * n * ((np.exp(bn)) / (bn) ** (2 * n)))

    x = gammaincinv(2. * n, g / gamma(2. * n))

    return (x / bn) ** n * r_eff


@custom_model
def sersic_enclosed_model(
        x,
        amplitude=1000,
        r_eff=30,
        n=2,
):
    return sersic_enclosed(x, amplitude, r_eff, n)


def petrosian_profile(r, r_eff, n):
    bn = gammaincinv(2. * n, 0.5)

    x = bn * (r / r_eff) ** (1 / n)

    g = gamma(2 * n) * gammainc(2 * n, x)

    return (np.exp(-x) * x ** (2 * n)) / (2 * n * g)


@custom_model
def petrosian_model(
    x, 
    r_eff=1, 
    n=4

):
    return petrosian_profile(x, r_eff, n)



from copy import deepcopy

import numpy as np

from photutils import detect_threshold
from photutils import deblend_sources
from photutils import detect_sources
from photutils.segmentation import SourceCatalog

from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from astropy.nddata import CCDData, Cutout2D

from matplotlib import pyplot as plt

__all__ = [
    'plot_segments', 'plot_segment_residual', 'get_source_position', 'get_source_elong',
    'get_source_ellip', 'get_source_theta', 'make_kernel', 'segm_mask', 'masked_segm_image',
    'make_segments', 'deblend_segments'
]


def plot_segments(segm, image=None, vmin=None, vmax=None, alpha=0.5, title=None):
    """
    Plot segmented areas over an image (2D array, if provided)
    """

    cmap = segm.make_cmap(seed=np.random.randint(1000000))

    if image is not None:
        plt.imshow(image, vmin=vmin, vmax=vmax, cmap="gist_gray")

    plt.imshow(segm, cmap=cmap, alpha=alpha)

    if title is not None:
        plt.title(title)

    plt.xlabel("Pixels")
    plt.ylabel("Pixels")


def plot_segment_residual(segm, image, vmin=None, vmax=None):
    """
    Plot segment subtracted image (residual)
    """
    temp = image.copy()
    temp[np.where(segm.data != 0)] = 0.0
    plt.imshow(temp, vmin=vmin, vmax=vmax)


def get_source_position(source):
    """Return max x, y value of a SourceCatalog or catalog row"""
    if isinstance(source, SourceCatalog):
        x, y = source.maxval_xindex, source.maxval_yindex
    else:
        x, y = source['maxval_xindex'], source['maxval_yindex']
    return x, y


def get_source_elong(source):
    """ Return SourceCatalog elongation"""
    return source.elongation.value if isinstance(source, SourceCatalog) else source['elongation']


def get_source_ellip(source):
    """ Return SourceCatalog ellipticity"""
    return source.ellipticity.value if isinstance(source, SourceCatalog) else source['ellipticity']


def get_source_theta(source):
    """ Return SourceCatalog orientation in rad"""
    return source.orientation.to('rad').value if isinstance(source, SourceCatalog) else np.deg2rad(source['orientation'])


def make_kernel(fwhm, kernel_size):
    sigma = fwhm * gaussian_fwhm_to_sigma
    kernel = Gaussian2DKernel(sigma, x_size=kernel_size, y_size=kernel_size)
    kernel.normalize()
    return kernel


def segm_mask(source, segm, mask_background=False):
    """
    Given a segmentation and a target with an label, returns a mask
    with all other sources masked in the original image.

    Parameters
    ----------

    source : int or photutils.segmentation.properties.SourceCatalog
        The catalog object for the target or label of target in the segmentation object.

    segm : photutils.segmentation.core.SegmentationImage
        Segmentation image.

    mask_background : bool
        Option to also mask out all the un-segmented background pixels.

    Returns
    -------

    mask : bool array
    """

    if isinstance(source, SourceCatalog):
        source = source.label

    mask = (segm.data == source)
    if not mask_background:
        mask = ((segm.data == 0) | mask)
    return mask


def masked_segm_image(source, image, segm, fill=None, mask_background=False):
    """
    Returns a masked image of the original image by masking out other sources

    Parameters
    ----------

    source : int or photutils.segmentation.properties.SourceCatalog
        The catalog object for the target or label of target in the segmentation object.

    image : CCDData or array
        Image to mask.

    segm : photutils.segmentation.core.SegmentationImage
        Segmentation image.

    fill : float
        Fill in the masked pixels with this value.

    mask_background : bool
        Option to also mask out all the un-segmented background pixels.

    Returns
    -------

    masked_image : CCDData or array
    """

    fill = np.nan if fill is None else fill
    mask = segm_mask(source, segm, mask_background)

    masked_image = deepcopy(image)

    if isinstance(masked_image, np.ndarray):
        masked_image[np.invert(mask)] = fill
    else:
        masked_image.data[np.invert(mask)] = fill

    return masked_image


def make_segments(image, npixels=None, threshold=3., fwhm=8., kernel_size=4):
    """
    Segment an image.

    Parameters
    ----------
    image : array like
        Input image

    npixels : int
        The number of connected pixels, each greater than ``threshold``,
        that an object must have to be detected.  ``npixels`` must be a
        positive integer.

    threshold : float or array-like
        The data value or pixel-wise data values to be used for the
        detection threshold. A 2D ``threshold`` must have the same shape
        as ``data``. See `~photutils.segmentation.detect_threshold` for
        one way to create a ``threshold`` image.

    fwhm : float
        FWHM of smoothing gaussian kernel.

    kernel_size : int
        Size of smoothing kernel.

    Returns
    -------

    segment_image : `~photutils.segmentation.SegmentationImage` or `None`
        A 2D segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values.  A
        value of zero is reserved for the background.  If no sources
        are found then `None` is returned.
    """

    if npixels is None:
        npixels = fwhm ** 2

    kernel = make_kernel(fwhm, kernel_size) if kernel_size else None

    return detect_sources(image, threshold, npixels=npixels, filter_kernel=kernel)


def deblend_segments(image, segm, npixels=None, fwhm=8., kernel_size=4, nlevels=30, contrast=1/1000):
    """
    Deblend overlapping sources labeled in a segmentation image.

    Parameters
    ----------
    image : array like
        Input image.

    segm : `~photutils.segmentation.SegmentationImage` or `None`
        A 2D segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values.  A
        value of zero is reserved for the background.  If no sources
        are found then `None` is returned.

    npixels : int
        The number of connected pixels, each greater than ``threshold``,
        that an object must have to be detected.  ``npixels`` must be a
        positive integer.

    fwhm : float
        FWHM of smoothing gaussian kernel.

    kernel_size : int
        Size of smoothing kernel.

    nlevels : int, optional
        The number of multi-thresholding levels to use.  Each source
        will be re-thresholded at ``nlevels`` levels spaced
        exponentially or linearly (see the ``mode`` keyword) between its
        minimum and maximum values within the source segment.

    contrast : float, optional
        The fraction of the total (blended) source flux that a local
        peak must have (at any one of the multi-thresholds) to be
        considered as a separate object.  ``contrast`` must be between 0
        and 1, inclusive.  If ``contrast = 0`` then every local peak
        will be made a separate object (maximum deblending).  If
        ``contrast = 1`` then no deblending will occur.  The default is
        0.001, which will deblend sources with a 7.5 magnitude
        difference.

    Returns
    -------
    segment_image : `~photutils.segmentation.SegmentationImage`
        A segmentation image, with the same shape as ``data``, where
        sources are marked by different positive integer values.  A
        value of zero is reserved for the background.
    """

    if npixels is None:
        npixels = fwhm ** 2

    kernel = make_kernel(fwhm, kernel_size) if kernel_size else None

    segm_deblend = deblend_sources(image, segm,
                                   npixels=npixels, filter_kernel=kernel,
                                   nlevels=nlevels, contrast=contrast)

    return segm_deblend


def make_catalog(image, threshold, wcs=None, deblend=True,
                 npixels=None, fwhm=8., kernel_size=4,
                 nlevels=30, contrast=1/1000,
                 plot=True, vmax=None, vmin=None):
    """
    This function constructs a catalog using `PhotUtils`. The `petrofit.segmentation.make_segments` and
    `petrofit.segmentation.deblend_segments` functions are used to construct segmentation maps and
    the resulting segmentation is turned into a vatalog (`photutils.segmentation.catalog.SourceCatalog`)

    Parameters
    ----------
    image : array like
        Input image.

    threshold : float or array-like
        The data value or pixel-wise data values to be used for the
        detection threshold. A 2D ``threshold`` must have the same shape
        as ``data``. See `~photutils.segmentation.detect_threshold` for
        one way to create a ``threshold`` image.

    wcs : astropy.wcs.WCS
        Astropy WCS for catalog to convert source locations to RA-Dec.

    deblend : bool
        Flag to enable deblending of sources in the segmentation map.

    npixels : int
        The number of connected pixels, each greater than ``threshold``,
        that an object must have to be detected.  ``npixels`` must be a
        positive integer.

    fwhm : float
        FWHM of smoothing gaussian kernel.

    kernel_size : int
        Size of smoothing kernel.

    nlevels : int, optional
        The number of multi-thresholding levels to use.  Each source
        will be re-thresholded at ``nlevels`` levels spaced
        exponentially or linearly (see the ``mode`` keyword) between its
        minimum and maximum values within the source segment.

    contrast : float, optional
        The fraction of the total (blended) source flux that a local
        peak must have (at any one of the multi-thresholds) to be
        considered as a separate object.  ``contrast`` must be between 0
        and 1, inclusive.  If ``contrast = 0`` then every local peak
        will be made a separate object (maximum deblending).  If
        ``contrast = 1`` then no deblending will occur.  The default is
        0.001, which will deblend sources with a 7.5 magnitude
        difference.

    plot : bool
        Flag to toggle plotting segmentation maps.

    vmin, vmax : float
        vmax and vmin values for plot.

    Returns
    -------
    cat, segm, segm_deblend
        cat : `photutils.segmentation.catalog.SourceCatalog`
            A catalog of sources

        segm : `photutils.segmentation.core.SegmentationImage`
            Segmentation map.

        segm_deblend : `photutils.segmentation.core.SegmentationImage`
            Deblended segmentation map.
    """

    if isinstance(image, CCDData) or isinstance(image, Cutout2D):
        image = image.data

    # Make segmentation map
    segm = make_segments(image,
                         threshold=threshold,
                         kernel_size=kernel_size,
                         fwhm=fwhm, npixels=npixels)

    if plot and segm:
        # Make plots
        if deblend:
            fig, ax = plt.subplots(1, 2)
            plt.sca(ax[0])
        plot_segments(segm, image=image, vmax=vmax, vmin=vmin, title='Segmentation Map')

    # Deblend segmentation map
    segm_deblend = None
    if deblend:
        segm_deblend = deblend_segments(image, segm,
                                        contrast=contrast,
                                        nlevels=nlevels,
                                        kernel_size=kernel_size,
                                        fwhm=fwhm, npixels=npixels)

        if plot and segm_deblend:
            # Make plots
            plt.sca(ax[1])
            plot_segments(segm_deblend, image=image, vmax=vmax, vmin=vmin, title='Deblended Segmentation Map')

    # Make catalog
    cat = SourceCatalog(image, segm_deblend if deblend else segm, wcs=wcs)

    return cat, segm, segm_deblend


import os
from sys import argv

def mount_bind_command(src, dst):
    return "--mount type=bind,src={},dst={}".format(src, dst)

if __name__ == '__main__':
    if len(argv) < 2:
        print('Command not supplied')
        print('python docker.py build')
        print('python docker.py run [extra_mount_paths]')
        print('python docker.py run_safe # (does not mount the host copy of petrofit dir)')

    # Build
    # -----
    # Command to build the Dockerfile
    elif argv[1] == 'build' or argv[1].replace("-", "") == "b":
        os.system('docker build -t petrofit .')

    # Mount and Run
    # -------------
    # Command to run the docker image in a container.
    # This will mount the host version of the repo so any changes to the code
    # in the host repo will be available in the docker container and the environment.
    # Thus the docker container will use the host's version of petrofit module since the
    # environment in the container is installed using `python setup.py develop`.
    # (you are able to update code on host without having to reinstall in the container)
    elif argv[1] == 'run' or argv[1].replace("-", "") == "r":
        path_to_this_dir = os.path.dirname(os.path.abspath(__file__))
        command = "docker run"
        command += " " + mount_bind_command(path_to_this_dir, "/home/jovyan/petrofit")

        # This section mounts (bind) any number of volumes on the host into the docker container.
        # The volumes are mounted into the `mount` folder inside the container.
        # Any edits there will also apply to the host version of the files.
        mounted_list = []
        if len(argv) > 2:
            for path in argv[2:]:
                src = os.path.abspath(path)
                dst = "/home/jovyan/mount/{}".format(os.path.basename(path))
                if dst in mounted_list:
                    raise Exception("Mount request includes multiple files or folders with the same name")
                mounted_list.append(dst)
                command += " " + mount_bind_command(src, dst)

        command += " " + "-it -p 8888:8888 petrofit"

        print(">>>", command, "\n")

        os.system(command)

    # Regular Run
    # -----------
    # Nothing is mounted on to the container.
    elif argv[1] == 'run_safe':
        os.system('docker run -it -p 8888:8888 petrofit')



import yaml

import numpy as np

from scipy.interpolate import interp1d

from matplotlib import pyplot as plt

from sklearn.tree import DecisionTreeRegressor

from .utils import closest_value_index, get_interpolated_values, pixel_to_angular
from .photometry import radial_elliptical_aperture

__all__ = [
    'plot_petrosian', 'calculate_petrosian', 'calculate_petrosian_r', 'discrete_petrosian_r',
    'calculate_r_total_flux', 'fraction_flux_to_r', 'calculate_r_half_light',
    'calculate_concentration_index', 'estimate_n', 'Petrosian', 'PetrosianCorrection'
]


def plot_petrosian(r_list, area_list, flux_list, epsilon=2., eta=0.2, plot_r=False):
    """
    Given photometric values, plots petrosian profile.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    area_list : numpy.array
        Array of aperture areas.

    flux_list : numpy.array
        Array of photometric flux values.

    epsilon : float
        Epsilon value (used to determine `r_total_flux`).
        N.B: `r_total_flux` = `r_petrosian` * `epsilon`

    eta : float, default=0.2
        Eta is the petrosian value which defines the `r_petrosian`.

    plot_r : bool
        If set to True, `r_half_light` and `r_total_flux` will be plotted.
    """
    petrosian_list = calculate_petrosian(area_list, flux_list)

    r_list_new, petrosian_list_new = get_interpolated_values(r_list[1:], petrosian_list[1:])

    plt.plot(r_list, petrosian_list, marker='o', linestyle='None', label='Data')
    plt.plot(r_list_new, petrosian_list_new, label='Interpolated [cubic]')

    r_petrosian = calculate_petrosian_r(r_list, area_list, flux_list, eta=eta)
    if not np.isnan(r_petrosian):
        plt.axvline(r_petrosian, linestyle='--', label="r_petrosian={:0.4f} pix".format(r_petrosian))
        plt.axhline(eta, linestyle='--', label='eta={:0.4f}'.format(eta))

    r_total = calculate_r_total_flux(r_list, area_list, flux_list, epsilon, eta)
    if not np.isnan(r_total) and plot_r:
        plt.axvline(r_total, linestyle='--', label="r_total={:0.4f} pix".format(r_total), c='black')

    r_half_light = calculate_r_half_light(r_list, flux_list, r_total)
    if not np.isnan(r_half_light) and plot_r:
        plt.axvline(r_half_light, linestyle='--', label="r_half_light={:0.4f} pix".format(r_half_light), c='gray')

    plt.legend(loc='best')

    plt.title("Petrosian")
    plt.xlabel("Aperture Radius [Pix]")
    plt.ylabel("Petrosian Value")


def calculate_petrosian(area_list, flux_list):
    """
    Given a list of aperture areas and associated fluxes, computes the petrosian curve.

    Parameters
    ----------
    area_list : numpy.array
        Array of aperture areas.

    flux_list : numpy.array
        Array of photometric flux values.

    Returns
    -------
    petrosian_list : numpy.array
        Array of petrosian values at each value of `area_list`
    """

    petrosian_list = []

    last_area = area_list[0]
    last_I = flux_list[0]
    petrosian_value = np.nan
    for i in range(len(area_list)):
        area = area_list[i]
        I = flux_list[i]

        if area != last_area:
            area_of_slice = area - last_area
            I_at_r = (I - last_I) / area_of_slice

            area_within_r = area
            I_avg_within_r = (I / area_within_r)

            petrosian_value = I_at_r / I_avg_within_r

        petrosian_list.append(petrosian_value)

        last_area = area
        last_I = I

    return np.array(petrosian_list)


def calculate_petrosian_r(r_list, area_list, flux_list, eta=0.2):
    """
    Calculate petrosian radius from photometric values using interpolation.
    The petrosian radius is defined as the radius at which the petrosian profile equals eta.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    area_list : numpy.array
        Array of aperture areas.

    flux_list : numpy.array
        Array of photometric flux values.

    eta : float, default=0.2
        Eta is the petrosian value which defines the `r_petrosian`.

    Returns
    -------
    r_petrosian : float or numpy.nan
        Petrosian radius
    """
    petrosian_list = calculate_petrosian(area_list, flux_list)

    # replace the first element with r=0 and petrosian=1
    # This is because the petrosian for the first radius in r_list can not be
    # computed, we instead replace that value with r=0, since that value is known to equal 0.
    r_list[0] = 0
    petrosian_list[0] = 1

    # Interpolate values
    r_list_new, petrosian_list_new = get_interpolated_values(r_list, petrosian_list)

    idx = closest_value_index(eta, petrosian_list_new)

    return np.nan if idx is None else r_list_new[idx]


def discrete_petrosian_r(r_list, area_list, flux_list, eta=0.2):
    """
    Calculate petrosian radius from photometric values by using the nearest value petrosian to eta.
    The petrosian radius is defined as the radius at which the petrosian profile equals eta.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    area_list : numpy.array
        Array of aperture areas.

    flux_list : numpy.array
        Array of photometric flux values.

    eta : float, default=0.2
        Eta is the petrosian value which defines the `r_petrosian`.

    Returns
    -------
    r_petrosian : float or numpy.nan
        Petrosian radius
    """

    petrosian_list = calculate_petrosian(area_list, flux_list)
    idx_list = np.where(petrosian_list <= eta)[0]

    r_petrosian = np.nan
    if idx_list.size > 0:
        idx = idx_list[0]
        r_petrosian = r_list[idx]

    return r_petrosian


def calculate_r_total_flux(r_list, area_list, flux_list, epsilon=2., eta=0.2, verbose=False):
    """
    Given photometric values, calculate Petrosian `r_total_flux`.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    area_list : numpy.array
        Array of aperture areas.

    flux_list : numpy.array
        Array of photometric flux values.

    epsilon : float
        Epsilon value (used to determine `r_total_flux`).
        N.B: `r_total_flux` = `r_petrosian` * `epsilon`

    eta : float, default=0.2
        Eta is the petrosian value which defines the `r_petrosian`.

    Returns
    -------
    r_total_flux : float or numpy.nan
        Total flux radius determined by the Petrosian profile.
    """

    r_petrosian = calculate_petrosian_r(r_list, area_list, flux_list, eta=eta)
    if np.isnan(r_petrosian):
        if verbose:
            print("r_petrosian could not be computed")
        return np.nan

    return r_petrosian * epsilon


def fraction_flux_to_r(r_list, flux_list, r_total_flux, fraction=0.5):
    """
    Given photometric values and `r_total_flux`, calculate radius which encloses a specified
    fraction of the total flux.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    flux_list : numpy.array
        Array of photometric flux values.

    r_total_flux : float
        Total flux radius (hint: `petrofit.petrosian.calculate_r_total_flux` can be used to measure this).

    fraction : float
        Fraction of flux total flux enclosed in target radius.

    Returns
    -------
    r_fraction_flux : float or numpy.nan
        Radius which encloses a specified fraction of the total flux (determined by the Petrosian profile).
    """

    if r_total_flux > max(r_list):
        return np.nan

    f = interp1d(r_list, flux_list, kind='cubic')
    total_flux = f(r_total_flux)
    fractional_flux = total_flux * fraction

    r_list_new, flux_list_new = get_interpolated_values(r_list, flux_list)

    # idx = abs(flux_list_new - fractional_flux).argmin()
    idx = closest_value_index(fractional_flux, flux_list_new, growing=True)
    return np.nan if idx is None else r_list_new[idx]


def calculate_r_half_light(r_list, flux_list, r_total_flux):
    """
    Given photometric values, calculate Petrosian `r_half_light`.

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    flux_list : numpy.array
        Array of photometric flux values.

    r_total_flux : float
        Total flux radius (hint: `petrofit.petrosian.calculate_r_total_flux` can be used to measure this).

    Returns
    -------
    r_half_light : float or numpy.nan
        Radius containing half of total flux.
    """
    return fraction_flux_to_r(r_list, flux_list, r_total_flux, fraction=0.5)


def calculate_concentration_index(r_list, flux_list, r_total_flux, fraction_1=0.2, fraction_2=0.8):
    """
    Calculates Petrosian concentration index.

    ``concentration_index = 5 * np.log10( r(fraction_2) / r(fraction_1) )``

    Parameters
    ----------
    r_list : numpy.array
        Array of radii in pixels.

    flux_list : numpy.array
        Array of photometric flux values.

    r_total_flux : float
        Total flux radius (hint: `petrofit.petrosian.calculate_r_total_flux` can be used to measure this).

    fraction_1 : float
        Fraction of total light enclosed by the radius in the denominator.

    fraction_2 : float
        Fraction of total light enclosed by the radius in the numerator.

    Returns
    -------
    r_fraction_1, r_fraction_2, concentration_index

        * r_fraction_1 : float or np.nan
            Radius containing `fraction_1` of the total flux.

        * r_fraction_2: float or np.nan
            Radius containing `fraction_2` of the total flux.

        * concentration_index : float or np.nan
            Concentration index
    """

    if r_total_flux > max(r_list):
        return [np.nan, np.nan, np.nan]

    r1 = fraction_flux_to_r(r_list, flux_list, r_total_flux, fraction=fraction_1)
    r2 = fraction_flux_to_r(r_list, flux_list, r_total_flux, fraction=fraction_2)

    # if np.any(np.isnan(np.array([r1, r2]))):
    #     return [np.nan, np.nan, np.nan]

    return r1, r2, 5 * np.log10(r2 / r1)


def estimate_n(c2080pet, verbose=False):
    n_list = [0.5, 0.75, 1, 1.5, 2, 4, 6, 8]
    c_pet_list = [2.14, 2.49, 2.78, 3.26, 3.63, 4.50, 4.99, 5.31]
    f = interp1d(c_pet_list, n_list, kind='cubic')
    try:
        return f(c2080pet)
    except ValueError:
        if verbose:
            print("Could not estimate n for {}, returning closest".format(c2080pet))
        return 0.5 if c2080pet < 2.14 else 5.31


class Petrosian:
    """
    Class that computes and plots Petrosian properties.
    """

    def __init__(self, r_list, area_list, flux_list,
                 epsilon=2., eta=0.2, verbose=False):
        """
        Initialize Petrosian properties by providing flux information.

        Parameters
        ----------
        r_list : numpy.array
            Array of radii in pixels.

        area_list : numpy.array
            Array of aperture areas.

        flux_list : numpy.array
            Array of photometric flux values.

        epsilon : float
            Epsilon value (used to determine `r_total_flux`).
            N.B: `r_total_flux` = `r_petrosian` * `epsilon`

        eta : float, default=0.2
            Eta is the petrosian value which defines the `r_petrosian`.

        verbose : bool
            Prints info using
        """

        self.verbose = verbose

        self.r_list = np.array(r_list)
        self.area_list = np.array(area_list)
        self.flux_list = np.array(flux_list)

        assert len(self.r_list) > 2, "At least 3 data points are needed to compute Petrosian."
        assert len(self.r_list.shape), "Input arrays should be one dimensional."
        assert self.r_list.shape == self.flux_list.shape == self.flux_list.shape

        self._epsilon = None
        self._eta = None

        self.epsilon = float(epsilon)
        self.eta = float(eta)

    @property
    def epsilon(self):
        """
        Epsilon value (used to determine `r_total_flux`).
        N.B: ``r_total_flux = r_petrosian * epsilon``
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        self._epsilon = value

    @property
    def eta(self):
        """Eta is the petrosian value which defines the `r_petrosian`"""
        return self._eta

    @eta.setter
    def eta(self, value):
        self._eta = value

    @property
    def r_petrosian(self):
        """
        The Petrosian radius is defined as the radius at which the petrosian profile equals eta.
        """
        return calculate_petrosian_r(self.r_list, self.area_list, self.flux_list,
                                     eta=self.eta)

    @property
    def r_total_flux(self):
        """The radius containing the total flux"""
        return calculate_r_total_flux(self.r_list, self.area_list, self.flux_list,
                                      epsilon=self.epsilon, eta=self.eta,
                                      verbose=self.verbose)

    @property
    def total_flux(self):
        """Returns the flux at `r_total_flux` or np.nan if out of range"""
        if not min(self.r_list) <= self.r_total_flux <= max(self.r_list):
            return np.nan

        f = interp1d(self.r_list, self.flux_list, kind='cubic')
        total_flux = f(self.r_total_flux)
        return float(total_flux)

    @property
    def r_half_light(self):
        """Radius containing half of the total Petrosian flux"""
        return calculate_r_half_light(self.r_list, self.flux_list, self.r_total_flux)

    @property
    def c2080(self):
        return self.concentration_index(fraction_1=0.2, fraction_2=0.8)[-1]

    @property
    def c5090(self):
        return self.concentration_index(fraction_1=0.5, fraction_2=0.9)[-1]

    def r_half_light_arcsec(self, wcs):
        """
        Given a wcs, compute the radius containing half of the total Petrosian flux in arcsec.
        """
        if not np.isnan(self.r_half_light):
            return pixel_to_angular(self.r_half_light, wcs).value
        return np.nan

    def r_total_flux_arcsec(self, wcs):
        """
        Given a wcs, compute the radius containing the total Petrosian flux in arcsec.
        """
        if not np.isnan(self.r_total_flux):
            return pixel_to_angular(self.r_total_flux, wcs).value
        return np.nan

    def concentration_index(self, fraction_1=0.2, fraction_2=0.8):
        """
        Calculates Petrosian concentration index.

        ``concentration_index = 5 * np.log10( r(fraction_2) / r(fraction_1) )``

        Parameters
        ----------
        fraction_1 : float
            Fraction of total light enclosed by the radius in the denominator.

        fraction_2 : float
            Fraction of total light enclosed by the radius in the numerator.

        Returns
        -------
        r_fraction_1, r_fraction_2, concentration_index

            * r_fraction_1 : float or np.nan
                Radius containing `fraction_1` of the total flux.

            * r_fraction_2: float or np.nan
                Radius containing `fraction_2` of the total flux.

            * concentration_index : float or np.nan
                Concentration index
        """
        return calculate_concentration_index(self.r_list, self.flux_list, self.r_total_flux,
                                             fraction_1=fraction_1, fraction_2=fraction_2)

    def fraction_flux_to_r(self, fraction=0.5):
        """Given a fraction, compute the radius containing that fraction of the Petrosian total flux"""
        return fraction_flux_to_r(self.r_list, self.flux_list, self.r_total_flux, fraction=fraction)

    def plot(self, plot_r=False, plot_normalized_flux=False):
        """
        Plots Petrosian profile.

        Parameters
        ----------
        plot_r : bool
            Plot total flux and half light radii.

        plot_normalized_flux:
            Over-plot the flux curve of growth by normalize the flux axis (max_flux=1).
        """
        plot_petrosian(self.r_list, self.area_list, self.flux_list, epsilon=self.epsilon, eta=self.eta, plot_r=plot_r)

        if plot_normalized_flux:
            plt.plot(self.r_list, self.flux_list/self.flux_list.max(), label='Normalized Flux', linestyle='--')

    def imshow(self, position=(0, 0), elong=1., theta=0., color=None, lw=None):
        """
        Make 2D plots of elliptical apertures with radii  of `r_half_light`, `r_total_flux`, `r_20` and `r_80`.

        Parameters
        ----------
        position : tuple
            (x, y) center of the apertures.

        elong : float
            Elongation of the aperture.

        theta : float
            The orientation of the aperture in rad.

        color : str
            Color override that will change the color of the apertures in the plot.

        lw : float
            Line width (thickness) of the plotted apertures.
        """

        labels = ['r_half_light', 'r_total_flux', 'r_20', 'r_80']
        radii = [self.r_half_light, self.r_total_flux, self.fraction_flux_to_r(.2), self.fraction_flux_to_r(.8)]
        colors = ['r', 'r', 'b', 'b']
        linestyles = ['dashed', 'solid', 'dotted', 'dashdot']

        for label, r, default_color, ls in zip(labels, radii, colors, linestyles):
            if not np.isnan(r) and r > 0:
                radial_elliptical_aperture(position, r, elong, theta).plot(
                    label=label, linestyle=ls, color=color if color else default_color, lw=lw)

        plt.scatter(*position, marker='+', color=color if color else 'red')


class PetrosianCorrection:
    """
    This class computes corrections for Petrosian given default Petrosian measurements.
    """

    def __init__(self, grid_file):
        """
        Parameters
        ----------
        grid_file : str
            Path to correction grid generated by `petrofit.correction.generate_petrosian_sersic_correction`
        """

        self._grid_file = grid_file
        self.grid = self._read_grid_file()
        self.regressor = self._train()

    def _read_grid_file(self):
        with open(self._grid_file, 'r') as infile:
            grid = yaml.load(infile, Loader=yaml.Loader)
        return grid

    def estimate_n(self, r_hl_pet, c2080pet, verbose=False):
        """
        Given the half light radius and c2080 computed using the default epsilon value,
        return an estimated sersic index n.
        """
        r_list = np.array(list(self.grid.keys()))
        r_eff = r_list[np.argmin(abs(r_list - r_hl_pet))]

        if verbose:
            print("r_eff = {}".format(r_eff))

        N_LIST = np.array(self.grid[r_eff]['n'])
        C_LIST = np.array(self.grid[r_eff]['c_index'])

        u, indices = np.unique(C_LIST, return_index=True)

        n_list = N_LIST[indices]
        c_pet_list = C_LIST[indices]
        f = interp1d(c_pet_list, n_list, kind='linear')
        try:
            return float(f(c2080pet))
        except ValueError:
            if verbose:
                print("Could not estimate n for {}, returning closest".format(c2080pet))
            return 0.5 if c2080pet < 2.14 else 5.31

    def estimate_epsilon(self, r_hl_pet, c2080pet, verbose=False):
        """
        Given the half light radius and c2080 computed using the default epsilon value,
        return a corrected epsilon value.
        """

        r_list = np.array(list(self.grid.keys()))
        r_eff = r_list[np.argmin(abs(r_list - r_hl_pet))]

        if verbose:
            print("r_eff = {}".format(r_eff))

        N_LIST = np.array(self.grid[r_eff]['epsilon'])
        C_LIST = np.array(self.grid[r_eff]['c_index'])

        u, indices = np.unique(C_LIST, return_index=True)
        indices = np.array(indices)

        n_list = N_LIST[indices]
        c_pet_list = C_LIST[indices]
        f = interp1d(c_pet_list, n_list, kind='linear')
        try:
            return float(f(c2080pet))
        except ValueError:
            if verbose:
                print("Could not estimate n for {}, returning closest".format(c2080pet))
            return 2

    # Experimental
    def _train(self):
        X = []
        N = []

        for r_eff in self.grid:
            for i in range(len(self.grid[r_eff]['n'])):
                c = self.grid[r_eff]['c_index'][i]
                n = self.grid[r_eff]['n'][i]
                r_p = self.grid[r_eff]['r_petrosian'][i]
                epsilon = self.grid[r_eff]['epsilon'][i]
                r_hl_petrosian = self.grid[r_eff]['r_hl_petrosian'][i]
                X.append(np.array([r_p, c]))
                N.append([n, epsilon])

        X = np.array(X)
        N = np.array(N)

        clf = DecisionTreeRegressor()
        clf.fit(X, N)
        return clf

    # Experimental
    def estimate_regressor(self, r_pet, c2080pet, verbose=False):
        """Experimental: Retrns n, epsilon using DecisionTreeRegressor"""
        return self.regressor.predict([[r_pet, c2080pet]])[0]



import os
import pytest
from copy import deepcopy

import numpy as np

from matplotlib import pyplot as plt

from petrofit.photometry import source_photometry, make_radius_list, order_cat

from petrofit.petrosian import Petrosian, PetrosianCorrection


def test_petrosian(sersic_petrosian):
    p = sersic_petrosian

    # Assert default
    assert p.epsilon == 2.
    assert p.eta == 0.2
    assert p.fraction_flux_to_r(0.5) == p.r_half_light
    assert p.fraction_flux_to_r(0.2) == p.concentration_index()[0]
    assert p.fraction_flux_to_r(0.8) == p.concentration_index()[1]

    # Check values
    assert (p.r_petrosian - 58.274255) / 58.274255 < 0.05
    assert (p.r_total_flux - 116.54851) / 116.54851 < 0.05
    assert (p.r_half_light - 24.233447) / 24.233447 < 0.05
    assert (p.total_flux - 9789.636481) / 9789.63648 < 0.05
    assert (p.concentration_index()[0] - 9.641128) / 9.641128 < 0.05

    # Test if plotting works
    p.plot(True, True)
    plt.show()

    p.imshow()
    plt.show()


def test_corrections(sersic_petrosian):
    p = sersic_petrosian

    pc = PetrosianCorrection(
        os.path.join(
            os.path.dirname(__file__),
            'concentration_index_grid_no_psf.yaml'
        )
    )

    c2080 = p.concentration_index()[-1]

    n = pc.estimate_n(p.r_half_light, c2080)

    n_err = (n - 2) / 2
    assert n_err < 0.05 # 5% error

    epsilon = pc.estimate_epsilon(p.r_half_light, c2080)

    assert (epsilon - 3.4) / 3.4 < 0.05

    corrected_p = Petrosian(
        p.r_list, p.area_list, p.flux_list,
        epsilon=epsilon
    )

    corrected_p.plot(True, True)
    plt.show()

    assert corrected_p.r_petrosian == p.r_petrosian
    assert (corrected_p.r_half_light - 25) / 25 < 0.05

    r_total_flux = p.r_total_flux

    p.epsilon = 4

    assert p.r_total_flux == r_total_flux * 2










import numpy as np

from astropy.modeling import models, fitting, Parameter
from astropy.modeling.optimizers import DEFAULT_ACC, DEFAULT_EPS

from matplotlib import pyplot as plt

__all__ = [
    'fit_model', 'model_to_image', 'fit_plane',
    'fit_gaussian2d', 'print_model_params', 'plot_fit'
]


def fit_model(image, model, maxiter=5000, epsilon=DEFAULT_EPS, acc=DEFAULT_ACC):
    """
    Wrapper function to conveniently fit an image to an input model.

    Parameters
    ----------
    image : array
        2D array to fit.

    model : `~astropy.modeling.FittableModel`
        AstroPy model to sample from.

    maxiter : int
        maximum number of iterations

    epsilon : float
        A suitable step length for the forward-difference
        approximation of the Jacobian (if model.fjac=None). If
        epsfcn is less than the machine precision, it is
        assumed that the relative errors in the functions are
        of the order of the machine precision.

    acc : float
        Relative error desired in the approximate solution

    Returns
    -------
    fitted_model, fitter

        * fitted_model : `~astropy.modeling.FittableModel`
            A copy of the input model with parameters set by the fitter.

        * fitter : LevMarLSQFitter
            Fitter used to estimate and set model parameters.
    """

    # Make x and y grid to fit to
    y_arange, x_arange = np.where(~(np.isnan(image)))

    z = image[(y_arange, x_arange)]

    # Fit model to grid
    fitter = fitting.LevMarLSQFitter()
    fitted_model= fitter(model, x_arange, y_arange, z, maxiter=maxiter, epsilon=epsilon, acc=acc)

    return fitted_model, fitter


def model_to_image(x, y, size, model):
    """
    Construct an image from a model.

    Parameters
    ----------
    x : int
        x center of sampling grid.

    y : int
        y center of sampling grid.

    size : int
        Size of sampling pixel grid.

    model : `~astropy.modeling.FittableModel`
        AstroPy model to sample from.

    Returns
    -------
    model_image : array
        2D image of the model.
    """
    y_arange, x_arange = np.mgrid[
                         int(y) - size//2:int(y) + size//2,
                         int(x) - size//2:int(x) + size//2, ]
    return model(x_arange, y_arange)


def fit_plane(image):
    """
    Given an image, fit a 2D plane.

    Parameters
    ----------
    image : array
        2D array to fit.

    Returns
    -------
    model : `~astropy.modeling.models.Planar2D`
        Plane model with best fit params
    """
    model = models.Planar2D(slope_x=0., slope_y=0, intercept=0)

    # Make x and y grid to fit to
    y_arange, x_arange = np.where(~(np.isnan(image)))

    z = image[(y_arange, x_arange)]

    # Fit model to grid
    fit = fitting.LinearLSQFitter()
    fitted_plane = fit(model, x_arange, y_arange, z)

    return fitted_plane, fit


def fit_gaussian2d(image):
    """
    Fit a 2D gaussian to a source in an image.

    Parameters
    ----------
    image : array
        2D array to fit.

    Returns
    -------
    fitted_model : `~astropy.modeling.models.Gaussian2D`
        AstroPy Gaussian2D model with parameters set by the fitter.
    """

    # Estimate center of target
    y_mean, x_mean = np.array(image.shape) // 2  # Center guess

    # Create model to fit
    model = models.Gaussian2D(amplitude=image.max(),
                              x_mean=x_mean,
                              y_mean=y_mean,
                              fixed={}
                              )

    # Fit model to grid
    fitted_model, fit = fit_model(image, model)

    return fitted_model


def print_model_params(model):
    """Print the params and values of an AstroPy model"""
    for param, value in zip(model.param_names, model.parameters):
        print("{:0.4f}\t{}".format(value,param))


def plot_fit(image, model, vmin=None, vmax=None):
    """
    Plot fitted model, its 1D fit profile and residuals.

    Parameters
    ----------
    image : array
        2D array that was fit by the model.

    model : `~astropy.modeling.FittableModel`
        Fitted AstroPy model.

    vmin : float
        Min plot value

    vmax : float
        Max plot value
    """
    if isinstance(model, models.Sersic2D):
        x_0, y_0 = model.x_0, model.y_0  # Center
    elif isinstance(model, models.Gaussian2D):
        x_0, y_0 = [i.value for i in [model.x_mean, model.y_mean]]
    else:
        x_0, y_0 = model.x_0, model.y_0  # Center

    if isinstance(x_0, Parameter):
        x_0, y_0 = [int(i.value) for i in [x_0, y_0]]

    fig = plt.figure(figsize=(12, 12))

    # Make x and y grid to plot to
    y_arange, x_arange = np.mgrid[:image.shape[0], :image.shape[1]]

    # Plot input image with FWHM and center
    # -------------------------------------
    ax0 = fig.add_subplot(221)

    ax0.imshow(image, vmin=vmin, vmax=vmax)
    ax0.axvline(x_0, label="Center")
    ax0.axhline(y_0)

    ax0.set_title("Image")
    ax0.set_xlabel("X Pixel")
    ax0.set_ylabel("Y Pixel")

    ax0.legend()

    # Plot residuals
    # ---------------

    residuals = image - model(x_arange, y_arange)
    # residuals[np.where(residuals < 0)] = 0.
    ax1 = fig.add_subplot(222)
    ax1.imshow(residuals, vmin=vmin, vmax=vmax)

    ax1.set_title("Residual (Image - Fit)")
    ax1.set_xlabel("X Pixel")
    ax1.set_ylabel("Y Pixel")

    # Prepare fine grid
    # -----------------

    # We need a fine grid to fill in inter-pixel values
    # Oversample by a factor of 10

    y_arange_fine, x_arange_fine = np.mgrid[:image.shape[0] * 10, :image.shape[1] * 10] / 10

    fine_image = model(x_arange_fine, y_arange_fine)
    x_slice_fine = fine_image[fine_image.shape[0] // 2, :]
    y_slice_fine = fine_image[:, fine_image.shape[1] // 2]

    # Plot X fit
    # ----------

    ax2 = fig.add_subplot(223)

    ax2.plot(x_arange_fine[1, :], x_slice_fine, c='r')
    ax2.scatter(x_arange[1, :], image[int(np.round(y_0)), :], c='black')

    ax2.set_title("X Cross Section")
    ax2.set_xlabel("X Pixel")
    ax2.set_ylabel("Flux")

    # Plot Y fit
    # ----------

    ax3 = fig.add_subplot(224)

    ax3.plot(y_arange_fine[:, 1], y_slice_fine, c='r')
    ax3.scatter(y_arange[:, 1], image[:, int(np.round(x_0))], c='black')

    ax3.set_title("Y Cross Section")
    ax3.set_xlabel("Y Pixel")
    ax3.set_ylabel("Flux")

    return fig, [ax0, ax1, ax2, ax3]




import numpy as np


from astropy.convolution import convolve
from astropy.modeling import FittableModel, Parameter, custom_model, models

from petrofit.models import Moffat2D, PSFModel, make_grid
from petrofit.fitting import model_to_image, fit_model
from petrofit.segmentation import masked_segm_image

import matplotlib.pyplot as plt


def test_psfmodel():
    """Test fitting and PSF convolution"""

    # Make model:

    imsize = 300

    sersic_model = models.Sersic2D(
        amplitude=1,
        r_eff=25,
        n=2,
        x_0=imsize / 2,
        y_0=imsize / 2,
        ellip=0,
        theta=0,
        bounds={
            'amplitude': (0., None),
            'r_eff': (0, None),
            'n': (0, 10),
            'ellip': (0, 1),
            'theta': (-2 * np.pi, 2 * np.pi),
        },
    )

    # Make model image
    image = model_to_image(imsize // 2, imsize // 2, imsize, sersic_model)

    # Make a PSF
    x_grid, y_grid = make_grid(51, 1)
    PSF = Moffat2D(x_0=25.0, y_0=25.0)(x_grid, y_grid)
    PSF /= PSF.sum()

    # Make a PSF image using model image and PSF
    psf_sersic_image = convolve(image, PSF)

    # Make a PSFModel
    psf_sersic_model = PSFModel.wrap(sersic_model, psf=PSF, oversample=None)
    psf_sersic_model.fixed['psf_pa'] = True

    # Make a PSFModel image
    psf_sersic_model_image = model_to_image(imsize // 2, imsize // 2, imsize, psf_sersic_model)

    # Compare the PSF image to PSFModel image
    error_arr =  abs(psf_sersic_model_image - psf_sersic_image) / psf_sersic_image
    assert np.max(error_arr) < 0.01 # max error less than 1% error

    # Test fitting

    # Change params with offset
    psf_sersic_model.r_eff = 23
    psf_sersic_model.x_0 = 0.6 + imsize / 2
    psf_sersic_model.y_0 = 0.1 + imsize / 2
    psf_sersic_model.n = 3


    # Fit
    fitted_model, fit_info = fit_model(
        psf_sersic_image, psf_sersic_model,
        maxiter=10000,
        epsilon=1.4901161193847656e-10,
        acc=1e-9,
    )

    # Generate a model image from the fitted model
    fitted_model_image = model_to_image(imsize // 2, imsize // 2, imsize, model=fitted_model)

    # Check if fit is close to actual
    error_arr = abs(fitted_model_image - psf_sersic_image) / psf_sersic_image
    assert np.max(error_arr) < 0.01 # max error less than 1% error



import os
import pytest

import numpy as np

from astropy.stats import sigma_clipped_stats
from astropy.nddata import CCDData
from astropy.modeling import models
from astropy.io import fits
from astropy.wcs import WCS


from petrofit.segmentation import make_catalog
from petrofit.petrosian import Petrosian
from petrofit.fitting import model_to_image
from petrofit.photometry import (
    make_radius_list,
    source_photometry,
    order_cat
)


def make_image():
    """Make 2D Sersic2D image"""
    imsize = 500

    d = 5

    sersic_model = models.Sersic2D(
        amplitude=1,
        r_eff=25,
        n=2,
        x_0=imsize / 2,
        y_0=imsize / 2,
        ellip=0,
        theta=0,
        bounds={
            'amplitude': (0., None),
            'r_eff': (0, None),
            'n': (0, 10),
            'ellip': (0, 1),
            'theta': (-2 * np.pi, 2 * np.pi),
        },
    )

    sersic_model += models.Sersic2D(
        amplitude=1,
        r_eff=25,
        n=2,
        x_0=imsize * 0.75,
        y_0=imsize * 0.75,
        ellip=0.5,
        theta=np.pi / 4,
        bounds={
            'amplitude': (0., None),
            'r_eff': (0, None),
            'n': (0, 10),
            'ellip': (0, 1),
            'theta': (-2 * np.pi, 2 * np.pi),
        },
    )

    sersic_model += models.Sersic2D(
        amplitude=1,
        r_eff=25,
        n=1,
        x_0=imsize * 0.25,
        y_0=imsize * 0.25,
        ellip=0.2,
        theta=np.pi / 6,
        bounds={
            'amplitude': (0., None),
            'r_eff': (0, None),
            'n': (0, 10),
            'ellip': (0, 1),
            'theta': (-2 * np.pi, 2 * np.pi),
        },
    )

    model_image = model_to_image(imsize // 2, imsize // 2, imsize, sersic_model)

    image_mean, image_median, image_stddev = sigma_clipped_stats(model_image, sigma=3)

    # model_image += np.random.normal(0, 3*image_stddev, size=model_image.shape)

    wcs = WCS()

    header = wcs.to_header()

    for param_name, param_val in zip(sersic_model.param_names, sersic_model.parameters):
        header[param_name] = param_val

    header['BUNIT'] = 'electron / s '

    path = "sersic_2d_image.fits.gz"
    sersic_2d_path = os.path.join(os.path.dirname(__file__), path)
    fits.writeto(
        sersic_2d_path,
        data=model_image,
        header=header,
        overwrite=True
    )



@pytest.fixture
def sersic_2d_image():
    """fixture for Sersic 2D image """
    path =  "sersic_2d_image.fits.gz"
    sersic_2d_path =  os.path.join(os.path.dirname(__file__), path)

    if not os.path.isfile(sersic_2d_path):
        make_image()

    return CCDData.read(sersic_2d_path)


@pytest.fixture
def segm_and_cat(sersic_2d_image):
    """fixture for segmentation and catalog"""
    image_mean, image_median, image_stddev = sigma_clipped_stats(sersic_2d_image, sigma=3)

    threshold = image_stddev * 3

    # Define smoothing kernel
    kernel_size = 3
    fwhm = 3

    # Min Source size (area)
    npixels = 4 ** 2

    return make_catalog(
        sersic_2d_image,
        threshold=threshold,
        deblend=True,
        kernel_size=kernel_size,
        fwhm=fwhm,
        npixels=npixels,
        contrast=0.00,
        plot=False,
    )



@pytest.fixture
def sersic_petrosian(sersic_2d_image, segm_and_cat):
    """fixture for Sersic model image Petrosian"""
    cat, segm, segm_deblend = segm_and_cat

    source = cat[order_cat(cat)[0]]

    max_pix = 200

    r_list = make_radius_list(
        max_pix=max_pix,  # Max pixel to go up to
        n=max_pix  # the number of radii to produce
    )

    flux_list, area_list, error_arr = source_photometry(

        # Inputs
        source,  # Source (`photutils.segmentation.catalog.SourceCatalog`)
        sersic_2d_image.data,  # Image as 2D array
        segm_deblend,  # Deblended segmentation map of image
        r_list,  # list of aperture radii

        # Options
        cutout_size=max(r_list) * 2,  # Cutout out size, set to double the max radius
        bkg_sub=False,  # Subtract background
        sigma=1, sigma_type='clip',  # Fit a 2D plane to pixels within 1 sigma of the mean
        plot=False, vmax=0, vmin=1,  # Show plot with max and min defined above
    )

    return Petrosian(r_list, area_list, flux_list)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *   # noqa
# ----------------------------------------------------------------------------

from .correction import *
from .fitting import *
from .models import *
from .petrosian import *
from .photometry import *
from .segmentation import *
from .utils import *



import pytest

from petrofit.segmentation import (
    plot_segments,
    plot_segment_residual,
    get_source_ellip,
    get_source_elong,
    get_source_position,
    get_source_theta,
)


def test_make_catalog(segm_and_cat):
    """This tests the segmentation, deblending and catalog code"""

    cat, segm, segm_deblend = segm_and_cat

    assert len(cat) == 3


def test_get_functions(segm_and_cat):
    """
    Test the convenience functions that
    get source properties
    """
    cat, segm, segm_deblend = segm_and_cat

    base_source = cat[0]

    source = base_source

    assert get_source_position(source) == (base_source.maxval_xindex, base_source.maxval_yindex)
    assert get_source_elong(source) == base_source.elongation.value
    assert get_source_ellip(source) == base_source.ellipticity.value
    assert get_source_theta(source) == base_source.orientation.to('rad').value


def test_plot_segments(sersic_2d_image, segm_and_cat):
    """Test segment plotting functions"""
    cat, segm, segm_deblend = segm_and_cat

    plot_segments(segm, vmax=1, vmin=0)

    plot_segments(segm_deblend, vmax=1, vmin=0)

    plot_segment_residual(segm, sersic_2d_image.data, vmax=1, vmin=0)





import warnings

import numpy as np

from astropy.nddata import Cutout2D
from astropy.stats import sigma_clipped_stats, sigma_clip
from astropy.utils.exceptions import AstropyWarning

from matplotlib import pyplot as plt

from photutils import  EllipticalAnnulus, EllipticalAperture

from .segmentation import masked_segm_image
from .fitting import fit_plane, model_to_image
from .segmentation import get_source_elong,  get_source_theta, get_source_position

__all__ = [
    'plot_apertures', 'flux_to_abmag', 'order_cat', 'radial_elliptical_aperture',
    'radial_elliptical_annulus', 'calculate_photometic_density', 'make_radius_list',
    'photometry_step', 'source_photometry'
]


def plot_apertures(image=None, apertures=[], vmin=None, vmax=None, color='white', lw=1.5):
    """
    Plot apertures on image

    Parameters
    ----------
    image : numpy.ndarray
        2D image array.

    apertures : list
        List of photutils Apertures.

    vmin, vmax : float
        vmax and vmin values for plot.

    color : string
        Matplotlib color for the apertures, default=White.

    lw : float
        Line width of aperture outline.
    """
    if image is not None:
        plt.imshow(image, cmap='Greys_r', vmin=vmin, vmax=vmax)

    for aperture in apertures:
        aperture.plot(axes=plt.gca(), color=color, lw=lw)

    if image or apertures:
        plt.title('Apertures')


def flux_to_abmag(flux, header):
    """Convert HST flux to AB Mag"""
    if not type(flux) in [int, float]:
        flux = np.array(flux)
        flux[np.where(flux <= 0)] = np.nan
    elif flux <= 0:
        return np.nan

    PHOTFLAM = header['PHOTFLAM']
    PHOTZPT = header['PHOTZPT']
    PHOTPLAM = header['PHOTPLAM']

    STMAG_ZPT = (-2.5 * np.log10(PHOTFLAM)) + PHOTZPT
    ABMAG_ZPT = STMAG_ZPT - (5. * np.log10(PHOTPLAM)) + 18.692

    return -2.5 * np.log10(flux) + ABMAG_ZPT


def order_cat(cat, key='area', reverse=True):
    """
    Sort a catalog by largest area and return the argsort

    Parameters
    ----------
    cat : `SourceCatalog` instance
        A `SourceCatalog` instance containing the properties of each source.

    key : string
        Key to sort.

    reverse : bool
        Reverse sorting order. Default is `True` to place largest values on top.

    Returns
    -------
    output : list
        A list of catalog indices ordered by largest area.
    """
    table = cat.to_table()[key]
    order_all = table.argsort()
    if reverse:
        return list(reversed(order_all))
    return order_all


def radial_elliptical_aperture(position, r, elong=1., theta=0.):
    """
    Helper function given a radius, elongation and theta,
    will make an elliptical aperture.

    Parameters
    ----------
    position : tuple
        (x, y) coords for center of aperture.

    r : int or float
        Semi-major radius of the aperture.

    elong : float
        Elongation.

    theta : float
        Orientation in rad.

    Returns
    -------
    EllipticalAperture
    """
    a, b = r, r / elong
    return EllipticalAperture(position, a, b, theta=theta)


def radial_elliptical_annulus(position, r, dr, elong=1., theta=0.):
    """
    Helper function given a radius, elongation and theta,
    will make an elliptical aperture.

    Parameters
    ----------
    position : tuple
        (x, y) coords for center of aperture

    r : int or float
        Semi-major radius of the inner ring

    dr : int or float
        Thickness of annulus (outer ring = r + dr).

    elong : float
        Elongation.

    theta : float
        Orientation in rad.

    Returns
    -------
    EllipticalAnnulus
    """

    a_in, b_in = r, r / elong
    a_out, b_out = r + dr, (r + dr) / elong

    return EllipticalAnnulus(position, a_in, a_out, b_out, theta=theta)


def calculate_photometic_density(r_list, flux_list, elong=1., theta=0.):
    """Compute value between radii"""
    density = []

    last_flux = 0
    last_area = 0
    for r, flux in zip(r_list, flux_list):
        aperture = radial_elliptical_aperture((0, 0), r, elong=elong, theta=theta)
        area = aperture.area
        density.append((flux - last_flux) / (area - last_area))
        last_area, last_flux = area, flux

    return density


def make_radius_list(max_pix, n):
    """Make an array of radii of size n up to max_pix"""
    r_list = [x * max_pix/n for x in range(1, n+1)]
    return np.array(r_list)


def photometry_step(position, r_list, image, error=None, mask=None, elong=1., theta=0.,
                    plot=False, vmin=0, vmax=None, method='exact'):
    """
    Core photometry function.  Given a position, a list of radii and the shape
    of apertures, calculate the photometry of the target in the image.

    Parameters
    ----------
    position : tuple
        (x, y) position in pixels.

    r_list : list
        A list of radii for apertures.

    image : 2D array
        Image to preform photometry on.

    error : 2D array
        Error map of the image.

    mask : 2D array
        Boolean array with True meaning that pixel is unmasked.

    elong : float
        Elongation.

    theta : float
        Orientation in rad.

    plot : bool
        Plot the target and apertures.

    vmin : int
        Min value for plot.

    vmax : int
        Max value for plot.

    method : {'exact', 'center', 'subpixel'}, optional
            The method used to determine the overlap of the aperture on
            the pixel grid.  Not all options are available for all
            aperture types.  Note that the more precise methods are
            generally slower.  The following methods are available:

                * ``'exact'`` (default):
                  The the exact fractional overlap of the aperture and
                  each pixel is calculated.  The returned mask will
                  contain values between 0 and 1.

                * ``'center'``:
                  A pixel is considered to be entirely in or out of the
                  aperture depending on whether its center is in or out
                  of the aperture.  The returned mask will contain
                  values only of 0 (out) and 1 (in).

                * ``'subpixel'``
                  A pixel is divided into subpixels (see the
                  ``subpixels`` keyword), each of which are considered
                  to be entirely in or out of the aperture depending on
                  whether its center is in or out of the aperture.  If
                  ``subpixels=1``, this method is equivalent to
                  ``'center'``.  The returned mask will contain values
                  between 0 and 1.

    Returns
    -------
    photometry, aperture_area, error
        Returns photometry, aperture area (unmasked pixels) and error at each radius.
    """

    flux_arr = []
    error_arr = []
    area_arr = []

    if plot:
        ax = plt.gca()
        plt.imshow(image, vmin=vmin, vmax=image.mean() * 10 if vmax is None else vmax)
        ax.set_title("Image and Aperture Radii")
        ax.set_xlabel("Pixels")
        ax.set_ylabel("Pixels")

    mask = ~mask if mask is not None else None
    for i, r in enumerate(r_list):
        aperture = radial_elliptical_aperture(position, r, elong=elong, theta=theta)

        photometric_value, photometric_err = aperture.do_photometry(data=image, error=error, mask=mask, method=method)
        aperture_area, aperture_area_err = aperture.do_photometry(data=np.ones_like(image), error=None,
                                                                  mask=mask, method=method)

        aperture_area = float(np.round(aperture_area, 6))
        photometric_value = float(np.round(photometric_value, 6))
        photometric_err = float(np.round(photometric_err, 6)) if photometric_err.size > 0 else np.nan

        if np.isnan(photometric_value):
            raise Exception("Nan photometric_value")

        if plot:
            aperture.plot(plt.gca(), color='w', alpha=0.5)


        flux_arr.append(photometric_value)
        area_arr.append(aperture_area)
        error_arr.append(photometric_err)

    return np.array(flux_arr), np.array(area_arr), np.array(error_arr)


def source_photometry(source, image, segm_deblend, r_list, error=None, cutout_size=None,
                      bkg_sub=False, sigma=3.0, sigma_type='clip', method='exact', mask_background=False,
                      plot=False, vmin=0, vmax=None, ):
    """
    Aperture photometry on a PhotUtils `SourceProperties`.

    Parameters
    ----------
    source : `photutils.segmentation.SourceProperties`
        `SourceProperties` (an entry in a `SourceCatalog`)

    image : 2D array
        Image to preform photometry on.

    segm_deblend : `SegmentationImage`
        Segmentation map of the image.

    r_list : list
        List of aperture radii.

    error : 2D array
        Error image (optional).

    cutout_size : int
        Size of cutout.

    bkg_sub : bool
        If the code should subtract the background using the `sigma` provided.

    sigma : float
        The sigma value used to determine noise pixels. Once the pixels above this value are masked,
        a 2D plane is fit to determine the background. The 2D plane model is then converted into an image and
        subtracted from the cutout of the target source. see the `sigma_type` on how this value will be used.

    sigma_type : {'clip', 'bound'}, optional
        The meaning of the provided sigma.
            * ``'clip'`` (default):
                Uses `astropy.stats.sigma_clipping.sigma_clip` to clip at the provided `sigma` std value.
                Note that `sigma` in this case is the number of stds above the mean.

            * ``'bound'``:
                After computing the mean of the image, clip at `mean - sigma` and `mean + sigma`.
                Note that `sigma` in this case is a value and not the number of stds above the mean.


    method : {'exact', 'center', 'subpixel'}, optional
        The method used to determine the overlap of the aperture on
        the pixel grid.  Not all options are available for all
        aperture types.  Note that the more precise methods are
        generally slower.  The following methods are available:

            * ``'exact'`` (default):
              The the exact fractional overlap of the aperture and
              each pixel is calculated.  The returned mask will
              contain values between 0 and 1.

            * ``'center'``:
              A pixel is considered to be entirely in or out of the
              aperture depending on whether its center is in or out
              of the aperture.  The returned mask will contain
              values only of 0 (out) and 1 (in).

            * ``'subpixel'``
              A pixel is divided into subpixels (see the
              ``subpixels`` keyword), each of which are considered
              to be entirely in or out of the aperture depending on
              whether its center is in or out of the aperture.  If
              ``subpixels=1``, this method is equivalent to
              ``'center'``.  The returned mask will contain values
              between 0 and 1.

    mask_background : bool
        Should background pixels, that are not part of any source in the segmentation map, be included?
        If False, only pixels inside the source's segmentation are unmasked.

    plot : bool
        Show plot of cutout and apertures.

    vmin : int
        Min value for plot.

    vmax : int
        Max value for plot.


    Returns
    -------

    flux_arr, area_arr, error_arr : (numpy.array, numpy.array, numpy.array)
        Tuple of arrays:

            * `flux_arr`: Photometric sum in aperture.

            * `area_arr`: Exact area of aperture.

            * `error_arr`: if error map is provided, error of measurements.
    """

    # Get source geometry
    # -------------------
    position = get_source_position(source)
    elong = get_source_elong(source)
    theta = get_source_theta(source)

    if cutout_size is None:
        cutout_size = np.ceil(max(r_list) * 3)

    cutout_size = int(cutout_size)
    if cutout_size % 2 == 1:
        cutout_size += 1

    # Error cutout
    # ------------
    masked_err = None
    if error is not None:
        masked_err = Cutout2D(error, position, cutout_size, mode='partial', fill_value=np.nan).data

    # Image Cutout
    # ------------
    full_masked_image = masked_segm_image(source, image, segm_deblend, fill=np.nan, mask_background=mask_background).data
    masked_nan_image = Cutout2D(full_masked_image, position, cutout_size, mode='partial', fill_value=np.nan)
    masked_image = masked_nan_image.data

    # Cutout for Stats
    # ----------------
    # This cutout has all sources masked
    stats_cutout_size = cutout_size  # max(source.bbox.ixmax - source.bbox.ixmin, source.bbox.iymax - source.bbox.iymin) * 2
    full_bg_image = masked_segm_image(0, image, segm_deblend, fill=np.nan, mask_background=False).data
    masked_stats_image = Cutout2D(full_bg_image, position, stats_cutout_size, mode='partial', fill_value=np.nan).data

    # Subtract Mean Plane
    # -------------------
    if bkg_sub:
        if len(np.where(~np.isnan(masked_stats_image))[0]) > 10:
            with warnings.catch_warnings():

                warnings.simplefilter('ignore', AstropyWarning)
                if sigma_type.lower() == 'clip':
                    fit_bg_image = masked_stats_image
                    fit_bg_image = sigma_clip(fit_bg_image, sigma)

                elif sigma_type.lower() == 'bound':
                    mean, median, std = sigma_clipped_stats(masked_stats_image, sigma=3,
                                                            mask=np.isnan(masked_stats_image.data))

                    fit_bg_image = masked_stats_image
                    fit_bg_image[np.where(fit_bg_image > mean + sigma)] = np.nan
                    fit_bg_image[np.where(fit_bg_image < mean - sigma)] = np.nan
                else:
                    raise ("background image masking sigma type not understood, try 'clip' or 'bound'")

                fitted_model, _ = fit_plane(fit_bg_image)

                # plt.imshow(fit_bg_image, vmax=vmax, vmin=vmin)
                # plt.show()
                # plt.imshow(model_to_image(cutout_size // 2, cutout_size // 2, cutout_size, fitted_model), vmin=vmin, vmax=vmax)
                # plt.show()
                # plt.imshow(masked_image, vmin=vmin, vmax=vmax)
                # plt.show()
                # plt.imshow(masked_image-model_to_image(cutout_size // 2, cutout_size // 2, cutout_size, fitted_model), vmin=vmin, vmax=vmax)
                # plt.show()

                masked_image -= model_to_image(cutout_size // 2, cutout_size // 2, cutout_size, fitted_model)
                if sigma_type.lower() == 'bound':
                    masked_image = np.clip(masked_image, - sigma, np.inf)

        elif plot:
            print("bkg_sub: Not enough datapoints, did not subtract.")

    # Make mask
    # ---------
    mask = np.ones_like(masked_image)
    mask[np.where(np.isnan(masked_image))] = 0
    mask = mask.astype(bool)

    position = np.array(masked_image.data.shape) / 2.

    if plot:
        print(source.label)
        fig, ax = plt.subplots(1, 2, figsize=[24, 12])

    if plot:
        plt.sca(ax[0])

    flux_arr, area_arr, error_arr = photometry_step(position, r_list, masked_image, error=masked_err, mask=mask,
                                                          elong=elong, theta=theta, plot=plot, vmin=vmin, vmax=vmax,
                                                          method=method)

    if plot:
        plt.sca(ax[1])
        plt.plot(r_list, flux_arr, c='black', linewidth=3)
        for r in r_list:
            plt.axvline(r, alpha=0.5, c='r')
        plt.title("Curve of Growth")
        plt.xlabel("Radius in Pixels")
        plt.ylabel("Flux Enclosed")
        plt.show()

        r = max(r_list)
        fig, ax = plt.subplots(1, 1, figsize=[24, 6])
        plt.plot(masked_image[:, int(position[0])], c='black', linewidth=3)
        plt.axhline(0, c='black')
        # plt.axhline(noise_sigma, c='b')
        plt.axvline(position[0], linestyle='--')
        plt.axvline(position[0] + r, alpha=0.5, c='r')
        plt.axvline(position[0] - r, alpha=0.5, c='r')
        plt.xlabel("Slice Along Y [pix]")
        plt.ylabel("Flux")

        fig, ax = plt.subplots(1, 1, figsize=[24, 6])

        plt.plot(masked_image[int(position[1]), :], c='black', linewidth=3)
        plt.axhline(0, c='black')
        # plt.axhline(noise_sigma, c='b')
        plt.axvline(position[0], linestyle='--')
        plt.axvline(position[0] + r, alpha=0.5, c='r')
        plt.axvline(position[0] - r, alpha=0.5, c='r')
        plt.xlabel("Slice Along X [pix]")
        plt.ylabel("Flux")

    return flux_arr, area_arr, error_arr


object_photometry = source_photometry  # Legacy



from copy import copy
import yaml

import numpy as np

from matplotlib import pyplot as plt

from scipy.interpolate import interp1d
from scipy.special import gammaincinv

from astropy.utils.console import ProgressBar
from astropy.modeling import models

from petrofit.models import PSFModel, sersic_enclosed, sersic_enclosed_inv
from petrofit.photometry import photometry_step
from petrofit.fitting import model_to_image
from petrofit.petrosian import Petrosian


__all__ = ['generate_petrosian_sersic_correction']


def generate_petrosian_sersic_correction(output_yaml_name, psf=None, r_eff_list=None, n_list=None,
                                         oversample=('x_0', 'y_0', 10, 10), plot=True):
    """
    This function generates corrections for Petrosian radii by simulating a galaxy image and measuring its properties.
    This is done to identify the correct `epsilon` value that, multiplied with `r_petrosian`, would give `r_total_flux`.
    To achieve this, an image created from a Sersic model and convolved with a PSF (if provided). The Petrosian radii
    and concentrations are computed using the default `epsilon` = 2. Since the real `r_total_flux` of the simulated galaxy
    is known, the correct `epsilon` can be determined by `epsilon = r_petrosian / r_total_flux`. The resulting grid will
    be used to map measured `r_petrosian` and `C2080` to the correct `epsilon` value. After the gird is computed, it
    will be saved to a yaml file which is readable by `petrofit.petrosian.PetrosianCorrection`.

    Parameters
    ----------
    output_yaml_name : str
        Name of output file, must be .yaml or .yml

    psf : numpy.array
        2D PSF image to pass to `petrofit.fitting.models.PSFModel`

    r_eff_list : list, (optional)
        List of `r_eff` (half light radii) in pixels to evaluate.

    n_list : list, (optional)
        List of Sersic indices to evaluate.

    oversample : int or tuple
        oversampling to pass to `petrofit.fitting.models.PSFModel`

    plot : bool
        Shows plot of photometry and Petrosian

    Returns
    -------
    petrosian_grid : dict
        Dictionary that is readable by `petrofit.petrosian.PetrosianCorrection`
        Also saves Yaml file that is readable by `petrofit.petrosian.PetrosianCorrection`.
    """

    if r_eff_list is None:
        r_eff_list = np.arange(10, 100 + 5, 5)

    if n_list is None:
        n_list = np.arange(1, 6.0 + 0.5, 0.5)

    petrosian_grid = {}

    with ProgressBar(len(r_eff_list) * len(n_list), ipython_widget=True) as bar:

        for r_eff_idx, r_eff in enumerate(r_eff_list):

            c_pet_list_output = []
            n_list_output = []
            r_p_list_output = []
            epsilon_list_output = []
            r_p_hl_list_output = []

            for n_idx, n in enumerate(n_list):
                bar.update()

                n = np.round(n, 2)
                amplitude = 100 / np.exp(gammaincinv(2. * n, 0.5))

                # Total flux
                total_flux = sersic_enclosed(
                    np.inf,
                    amplitude=amplitude,
                    r_eff=r_eff,
                    n=n
                )
                total_flux = total_flux * 0.99

                # Total flux radius
                r_total_flux = sersic_enclosed_inv(
                    total_flux,
                    amplitude=amplitude,
                    r_eff=r_eff,
                    n=n
                )
                ori_r_total_flux = r_total_flux
                max_r = r_total_flux * 2 if n < 2 else r_total_flux * 1.2

                # Make r_list
                if max_r >= 200:
                    r_list = [x for x in range(1, 201, 2)]
                    r_list += [x for x in range(300, int(max_r) + 100, 100)]
                else:
                    r_list = [x for x in range(1, int(max_r) + 2, 2)]
                r_list = np.array(r_list)

                image_size = max(r_list) * 2

                x_0 = image_size // 2
                y_0 = image_size // 2

                # Define model
                galaxy_model = models.Sersic2D(
                    amplitude=amplitude,
                    r_eff=r_eff,
                    n=n,
                    x_0=x_0,
                    y_0=y_0,
                    ellip=0.,
                    theta=0.,
                )

                # PSF weap
                galaxy_model = PSFModel.wrap(galaxy_model, psf=psf, oversample=oversample)

                galaxy_image = model_to_image(x_0, y_0, image_size, galaxy_model)

                flux_list, area_list, err = photometry_step((image_size // 2, image_size // 2), r_list, galaxy_image,
                                                            plot=plot,
                                                            vmax=amplitude / 100)
                plt.show()

                if plot:
                    print(r_eff, n, r_total_flux)
                    plt.show()

                # Petrosian from Photometry
                p = Petrosian(r_list, area_list, flux_list)
                rc1, rc2, c_index = p.concentration_index()
                if np.any(np.isnan(np.array([rc1, rc2, c_index]))):
                    raise Exception("concentration_index cant be computed")

                # Compute new r_total_flux
                u, indices = np.unique(flux_list, return_index=True)
                indices = np.array(indices)
                f = interp1d(flux_list[indices], r_list[indices], kind='linear')
                corrected_r_total_flux = f(total_flux)

                # Compute corrections
                corrected_epsilon = corrected_r_total_flux / p.r_petrosian
                corrected_p = copy(p)
                corrected_p.epsilon = corrected_epsilon
                if plot:
                    corrected_p.plot(True, True)
                    plt.show()
                    print(corrected_epsilon)
                    print(r_eff, p.r_half_light, corrected_p.r_half_light)
                    print(ori_r_total_flux, p.r_total_flux, corrected_p.r_total_flux, corrected_r_total_flux)
                    print(" ")

                c_pet_list_output.append(p.concentration_index()[-1])
                n_list_output.append(n)
                r_p_list_output.append(p.r_petrosian)
                epsilon_list_output.append(corrected_epsilon)
                r_p_hl_list_output.append(corrected_p.r_half_light)

                del galaxy_model, galaxy_image

            petrosian_grid[r_eff] = {
                'c_index': c_pet_list_output,
                'n': n_list_output,
                'epsilon': epsilon_list_output,

                'r_petrosian': r_p_list_output,
                'r_hl_petrosian': r_p_hl_list_output
            }

    if output_yaml_name is not None:
        with open(output_yaml_name, 'w') as outfile:
            print(outfile.name)
            yaml.dump(petrosian_grid, outfile)

    return petrosian_grid


"""Configure Test Suite.

This file is used to configure the behavior of pytest when using the Astropy
test infrastructure. It needs to live inside the package in order for it to
get picked up when running the tests inside an interpreter using
packagename.test

"""

import os

from astropy.version import version as astropy_version

# For Astropy 3.0 and later, we can use the standalone pytest plugin
if astropy_version < '3.0':
    from astropy.tests.pytest_plugins import *  # noqa
    del pytest_report_header
    ASTROPY_HEADER = True
else:
    try:
        from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
        ASTROPY_HEADER = True
    except ImportError:
        ASTROPY_HEADER = False


def pytest_configure(config):
    """Configure Pytest with Astropy.

    Parameters
    ----------
    config : pytest configuration

    """
    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'

        from . import __version__
        packagename = os.path.basename(os.path.dirname(__file__))
        TESTED_VERSIONS[packagename] = __version__

# Uncomment the last two lines in this block to treat all DeprecationWarnings as
# exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()



import re

import numpy as np

from scipy.interpolate import interp1d

from astropy.stats import gaussian_sigma_to_fwhm
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy import units as u
from astropy.coordinates import SkyCoord

from matplotlib import pyplot as plt

from .fitting import fit_gaussian2d, plot_fit

__all__ = [
    'match_catalogs', 'angular_to_pixel', 'pixel_to_angular',
    'elliptical_area_to_r', 'circle_area_to_r', 'get_interpolated_values',
    'closest_value_index', 'plot_target', 'cutout', 'cutout_subtract',
    'measure_fwhm'
]

def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [ convert(c) for c in re.split('([0-9]+)', key) ]
    return sorted(l, key = alphanum_key)


def match_catalogs(ra_1, dec_1, ra_2, dec_2, unit='deg'):
    """Wrapper for `SkyCoord.match_to_catalog_sky`"""
    cat1_coords = SkyCoord(ra=ra_1, dec=dec_1, unit=unit)
    cat2_coords = SkyCoord(ra=ra_2, dec=dec_2, unit=unit)
    return cat1_coords.match_to_catalog_sky(cat2_coords)


def angular_to_pixel(angular_diameter, wcs):
    pixel_scales = proj_plane_pixel_scales(wcs)
    assert np.allclose(*pixel_scales)
    pixel_scale = pixel_scales[0] * wcs.wcs.cunit[0] / u.pix

    pixel_size = angular_diameter / pixel_scale.to(angular_diameter.unit / u.pix)
    pixel_size = pixel_size.value

    return pixel_size


def pixel_to_angular(pixel_size, wcs):
    pixel_scales = proj_plane_pixel_scales(wcs)
    assert np.allclose(*pixel_scales)
    pixel_scale = pixel_scales[0] * wcs.wcs.cunit[0] / u.pix

    if not hasattr(pixel_size, 'unit'):
        pixel_size = pixel_size * u.pix

    angular_diameter = pixel_size * pixel_scale.to(u.arcsec / u.pix)
    return angular_diameter


def elliptical_area_to_r(area, elong):
    a = np.sqrt(elong * area / (np.pi))
    b = a / elong
    return a, b


def circle_area_to_r(area):
    return np.sqrt(area / (np.pi))


def get_interpolated_values(x, y, num=5000, kind='cubic'):
    f = interp1d(x, y, kind=kind)
    x_new = np.linspace(min(x), max(x), num=num, endpoint=True)
    y_new = f(x_new)
    return x_new, y_new


def closest_value_index(value, array, growing=False):
    """Return first index closes to value"""

    if not growing:
        idx_list = np.where(array <= value)[0]
    elif growing:
        idx_list = np.where(array >= value)[0]

    idx = None
    if idx_list.size > 0:
        idx = idx_list[0]
        idx = abs(array[:idx + 1] - value).argmin()
    return idx


def plot_target(position, image, size, vmin=None, vmax=None):
    x, y = position
    if not isinstance(image, np.ndarray):
        image = image.data
    plt.imshow(image, vmin=vmin, vmax=vmax)
    plt.plot(x, y, '+', c='r', label='Target')
    plt.xlim(x-size, x+size)
    plt.ylim(y-size, y+size)


def cutout(image, x, y, dx, dy=None, vmin=None, vmax=None):
    """
    Clip and make a cutout of an image

    Parameters
    ----------
    image : array like
        Input image.
    x, y : int
        Center of cutout. The indexing is array[y, x] and
        x is the x axis when plotted.
    dx, dy : int
        Size of image in x and y direction. The indexing is array[y, x] and
        x is the x axis when plotted. If dy is not provided it will be set
        to the same size as dx.
    vmin, vmax : float
        max and min values to clip the input image.


    Returns
    -------
    Copied array
        Clipped and cropped image.
    """

    if dy is None:
        dy = dx

    vmin = vmin if vmin else image.min()
    vmax = vmax if vmax else image.max()
    image = np.clip(image, vmin, vmax)

    bounds = np.array([y - dy // 2, y + dy // 2, x - dx // 2, x + dx // 2])
    bounds[bounds < 0] = 0
    ymin, ymax, xmin, xmax = bounds

    return image[ymin:ymax, xmin:xmax].copy()


def cutout_subtract(image, target, x, y):
    """
    Subtract cutout from image

    Parameters
    ----------
    image : array like
        Main image

    target : array like
        Cutout image

    x, y : int
        Center to subtract from

    Returns
    -------
    Copied array
        subtracted
    """

    dy, dx = target.shape
    bounds = np.array([y - dy // 2, y + dy // 2, x - dx // 2, x + dx // 2])
    bounds[bounds < 0] = 0
    ymin, ymax, xmin, xmax = bounds
    image[ymin:ymax, xmin:xmax] -= target
    return image


def measure_fwhm(image, plot=True, printout=True):
    """
    Find the 2D FWHM of a background/continuum subtracted cutout image of a target.
    The target should be centered and cropped in the cutout.
    Use lcbg.utils.cutout for cropping targets.
    FWHM is estimated using the sigmas from a 2D gaussian fit of the target's flux.
    The FWHM is returned as a tuple of the FWHM in the x and y directions.

    Parameters
    ----------
    image : array like
        Input background/continuum subtracted cutout image.
    printout : bool
        Print out info.
    plot : bool
        To plot fit or not.

    Returns
    -------
    tuple : array of floats
        FWHM in x and y directions.
    """

    # Find FWHM
    # ----------

    fitted_line = fit_gaussian2d(image)

    # Find fitted center
    x_mean, y_mean = [i.value for i in [fitted_line.x_mean, fitted_line.y_mean]]

    # Estimate FWHM using gaussian_sigma_to_fwhm
    x_fwhm = fitted_line.x_stddev * gaussian_sigma_to_fwhm
    y_fwhm = fitted_line.y_stddev * gaussian_sigma_to_fwhm

    # Find half max
    hm = fitted_line(x_mean, y_mean) / 2.

    # Find the mean of the x and y direction
    mean_fwhm = np.mean([x_fwhm, y_fwhm])
    mean_fwhm = int(np.round(mean_fwhm))

    # Print info about fit and FWHM
    # ------------------------------

    if printout:
        print("Image Max: {}".format(image.max()))
        print("Amplitude: {}".format(fitted_line.amplitude.value))
        print("Center: ({}, {})".format(x_mean, y_mean))
        print("Sigma = ({}, {})".format(fitted_line.x_stddev.value,
                                        fitted_line.y_stddev.value, ))

        print("Mean FWHM: {} Pix ".format(mean_fwhm))
        print("FWHM: (x={}, y={}) Pix ".format(x_fwhm, y_fwhm))

    if plot:

        fig, [ax0, ax1, ax2, ax3] = plot_fit(image, fitted_line)

        # Make x and y grid to plot to
        y_arange, x_arange = np.mgrid[:image.shape[0], :image.shape[1]]

        # Plot input image with FWHM and center
        # -------------------------------------

        ax0.imshow(image, cmap='gray_r')

        ax0.axvline(x_mean - x_fwhm / 2, c='c', linestyle="--", label="X FWHM")
        ax0.axvline(x_mean + x_fwhm / 2, c='c', linestyle="--")

        ax0.axhline(y_mean - y_fwhm / 2, c='g', linestyle="--", label="Y FWHM")
        ax0.axhline(y_mean + y_fwhm / 2, c='g', linestyle="--")

        ax0.set_title("Center and FWHM Plot")
        ax0.legend()

        # Plot X fit
        # ----------

        ax2.axvline(x_mean, linestyle="-", label="Center")
        ax2.axvline(x_mean - x_fwhm / 2, c='c', linestyle="--", label="X FWHM")
        ax2.axvline(x_mean + x_fwhm / 2, c='c', linestyle="--")
        ax2.axhline(hm, c="black", linestyle="--", label="Half Max")

        ax2.legend()

        # Plot Y fit
        # ----------

        ax3.axvline(y_mean, linestyle="-", label="Center")
        ax3.axvline(y_mean - y_fwhm / 2, c='g', linestyle="--", label="Y FWHM")
        ax3.axvline(y_mean + y_fwhm / 2, c='g', linestyle="--")
        ax3.axhline(hm, c="black", linestyle="--", label="Half Max")

        ax3.legend()

        plt.show()

    return np.array([x_fwhm, y_fwhm])









import numpy as np

from astropy.io import fits

from petrofit.photometry import (
    radial_elliptical_aperture,
    radial_elliptical_annulus,
    make_radius_list,
    photometry_step,
    source_photometry,
    flux_to_abmag,
    order_cat
)


def test_radial_elliptical_aperture():
    """Test if loading elliptical aperture"""
    radial_elliptical_aperture(position=(0, 0), r=10)


def test_radial_elliptical_annulus():
    """Test if loading elliptical annulus"""
    radial_elliptical_annulus(position=(0, 0), r=10, dr=5)


def test_make_radius_list():
    """test the make_radius_list function"""
    # Simple r_list
    max_pix = 100
    n = 50

    r_list = make_radius_list(max_pix, n)

    assert max(r_list) == max_pix
    assert len(r_list) == n
    assert not np.any(r_list <= 0)

    # Odd division r_list
    max_pix = 100
    n = 30

    r_list = make_radius_list(max_pix, n)

    assert max(r_list) == max_pix
    assert len(r_list) == n
    assert not np.any(r_list <= 0)


def test_photometry_step():
    """
    Test photometry core by doing photometry
    where the total flux = pixel area.
    """
    image_size = 30
    ones_image = np.ones((30, 30))

    r_list = make_radius_list(image_size // 2, image_size // 2)

    results = photometry_step(
        np.array(ones_image.shape) // 2,
        r_list,
        ones_image,
        plot=True
    )

    results = [r_list] + list(results)

    results = [np.round(l, 6) for l in results]

    for r, photometry, aperture_area, error in zip(*results):
        area = np.round(np.pi * r ** 2, 6)

        photo_error = abs(photometry - area) / area
        assert photo_error < 0.01, "{}".format(photo_error)

        area_error = abs(aperture_area - area) / area
        assert area_error < 0.01, "{}".format(area_error)


def test_source_photometry(sersic_2d_image, segm_and_cat):
    """
    Test the source_photometry function by doing photometry on
    a simulated image.
    """
    cat, segm, segm_deblend = segm_and_cat

    source = cat[0]

    max_pix = 120

    r_list = make_radius_list(
        max_pix=max_pix, # Max pixel to go up to
        n=max_pix # the number of radii to produce
    )


    flux_arr, area_arr, error_arr = source_photometry(

        # Inputs
        source, # Source (`photutils.segmentation.catalog.SourceCatalog`)
        sersic_2d_image.data, # Image as 2D array
        segm_deblend, # Deblended segmentation map of image
        r_list, # list of aperture radii

        # Options
        cutout_size=max(r_list)*2, # Cutout out size, set to double the max radius
        bkg_sub=True, # Subtract background
        sigma=1, sigma_type='clip', # Fit a 2D plane to pixels within 1 sigma of the mean
        plot=True, vmax=0, vmin=1, # Show plot with max and min defined above
    )



def test_order_cat(segm_and_cat):
    """Test if ordering is correct"""
    cat, segm, segm_deblend = segm_and_cat
    order = order_cat(cat, 'area')

    source = None
    for idx in order:
        if source is not None:
            assert source.area > cat[idx].area

        source = cat[idx]


def test_flux_to_abmag():
    """Test HST Flux to AB mag conversions"""

    header = fits.Header()

    header['PHOTFLAM'] = 3.03865739999999e-20
    header['PHOTZPT'] =  -21.10000000000001
    header['PHOTPLAM'] = 10551.04700000001

    mags = np.round(flux_to_abmag(100, header), 2)

    assert mags == 21.27
